package org.expeditee.items.widgets;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JToolBar;

import org.expeditee.gui.FreeItems;
import org.expeditee.gui.MessageBay;
import org.expeditee.gui.UserSettings;
import org.expeditee.items.Text;

/**
 * Web browser using a JavaFX WebView.
 * 
 * @author ngw8
 * @author jts21
 */
public class JfxBrowser extends DataFrameWidget {
	
	private static final String BACK = "back";
	private static final String FORWARD = "forward";
	private static final String REFRESH = "refresh";
	
	private WebBrowserPanel _browser;

	static Class JFXPanel;

	static Class Platform;
	static Method PlatformRunLater;

	static Class Group;
	static Method GroupGetChildren;

	static Class Scene;
	static Constructor SceneConstructor;

	static Class WebView;
	static Method JFXPanelSetScene;
	static Method WebViewSetPrefSize;
	static Method WebViewGetEngine;

	static Class WebEngine;
	static Method WebEngineLoad;
	static Method WebEngineGetLocation;
	static Method WebEngineGetLoadWorker;
	static Method WebEngineGetHistory;
	static Method WebEngineReload;

	static Class WebHistory;
	static Method WebHistoryGo;

	static Class Worker;
	static Method WorkerStateProperty;

	static Class ReadOnlyObjectProperty;
	static Method ReadOnlyObjectPropertyAddListener;

	static Class ChangeListener;

	static Class ObservableValue;

	static Class State;
	static List<Object> StateConstants;

	static {
		String javaLibDir = System.getProperty("java.home") + File.separator + "lib" + File.separator;
		try {
			ClassLoader classLoader = ClassLoader.getSystemClassLoader();

			File jar = new File(javaLibDir + "jfxrt.jar");

			Method addURL = URLClassLoader.class.getDeclaredMethod("addURL", URL.class);
			addURL.setAccessible(true);
			addURL.invoke(classLoader, jar.toURI().toURL());

			JFXPanel = Class.forName("javafx.embed.swing.JFXPanel", false, classLoader);

			Platform = classLoader.loadClass("javafx.application.Platform");
			PlatformRunLater = Platform.getMethod("runLater", Runnable.class);

			// Setting JFX to not exit when all panels are hidden, otherwise moving, resizing and changing frames will
			// cause JFX to exit, preventing JFX elements from displaying at all
			Platform.getMethod("setImplicitExit", boolean.class).invoke(null, false);

			Group = classLoader.loadClass("javafx.scene.Group");
			GroupGetChildren = Group.getMethod("getChildren");

			Scene = classLoader.loadClass("javafx.scene.Scene");
			SceneConstructor = Scene.getConstructor(classLoader.loadClass("javafx.scene.Parent"));

			WebView = classLoader.loadClass("javafx.scene.web.WebView");
			JFXPanelSetScene = JFXPanel.getMethod("setScene", Scene);
			WebViewSetPrefSize = WebView.getMethod("setPrefSize", double.class, double.class);
			WebViewGetEngine = WebView.getMethod("getEngine");

			WebEngine = classLoader.loadClass("javafx.scene.web.WebEngine");
			WebEngineLoad = WebEngine.getMethod("load", String.class);
			WebEngineGetLocation = WebEngine.getMethod("getLocation");
			WebEngineGetLoadWorker = WebEngine.getMethod("getLoadWorker");
			WebEngineGetHistory = WebEngine.getMethod("getHistory");
			WebEngineReload = WebEngine.getMethod("reload");
			
			WebHistory = classLoader.loadClass("javafx.scene.web.WebHistory");
			WebHistoryGo = WebHistory.getMethod("go", int.class);

			Worker = classLoader.loadClass("javafx.concurrent.Worker");
			WorkerStateProperty = Worker.getMethod("stateProperty");

			ChangeListener = classLoader.loadClass("javafx.beans.value.ChangeListener");

			ReadOnlyObjectProperty = classLoader.loadClass("javafx.beans.property.ReadOnlyObjectProperty");
			ReadOnlyObjectPropertyAddListener = ReadOnlyObjectProperty.getMethod("addListener", ChangeListener);

			ObservableValue = classLoader.loadClass("javafx.beans.value.ObservableValue");

			State = classLoader.loadClass("javafx.concurrent.Worker$State");
			StateConstants = Arrays.asList(State.getEnumConstants());

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private static class WebBrowserPanel extends JPanel implements ComponentListener {

		private static final long serialVersionUID = 1L;

		Object webview;
		Object jfxPanel;
		JfxBrowser owner;

		private JTextField urlField;

		/**
		 * @return A JPanel that either contains a JavaFX Webview or an empty panel if the JFX panel can't be
		 *         initialized
		 */
		public WebBrowserPanel(final String url) {

			super(new BorderLayout());
			this.addComponentListener(this);
			try {
				jfxPanel = JFXPanel.newInstance();

				// Toolbar (that can't be dragged) to hold the nav buttons, url bar, etc
				JToolBar toolBar = new JToolBar();
				toolBar.setFloatable(false);


				JButton backButton = new JButton("Back");
				toolBar.add(backButton);

				JButton forwardButton = new JButton("Forward");
				toolBar.add(forwardButton);

				urlField = new JTextField(url);
				toolBar.add(urlField);

				backButton.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent arg0) {
						owner.navigateForward();
					}
				});

				forwardButton.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent arg0) {
						owner.navigateBack();
					}
				});

				add(toolBar, BorderLayout.PAGE_START);

				this.add((Component) jfxPanel);

				// Setting bg color to be the same as bg color of widgets when they're being dragged/resized
				this.setBackground(new Color(100, 100, 100));

				PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						initFx(url);
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		/**
		 * Sets up the browser frame. JFX requires this to be run on a new thread.
		 * 
		 * @param url
		 *            The URL to be loaded when the browser is created
		 */
		private void initFx(String url) {
			try {
				this.webview = WebView.newInstance();

				Object scene = SceneConstructor.newInstance(this.webview);

				JFXPanelSetScene.invoke(jfxPanel, scene);

				ReadOnlyObjectPropertyAddListener.invoke(WorkerStateProperty.invoke(WebEngineGetLoadWorker
				        .invoke(WebViewGetEngine.invoke(this.webview))), java.lang.reflect.Proxy.newProxyInstance(
				        ChangeListener.getClassLoader(), new java.lang.Class[] { ChangeListener },
				        new java.lang.reflect.InvocationHandler() {
					        @Override
					        public Object invoke(Object proxy, java.lang.reflect.Method method, Object[] args)
					                throws java.lang.Throwable {
						        String method_name = method.getName();
						        // Class<?>[] classes = method.getParameterTypes();
						        // public void changed(ObservableValue ov, State oldState, State newState)
						        if (method_name.equals("changed")) {
							        // changed takes 3 args
							        if (args == null || args.length != 3) {
								        return null;
							        }
							        // args[0] is the ObservableValue
							        System.out.println(args[0]);
							        System.out.println(args[0].getClass());
							        // args[2] is the new State
							        if (args[2].getClass() == State) {
								        int id = StateConstants.indexOf(args[2]);
								        switch (id) {
										case 0: // READY
											MessageBay.displayMessage("WebEngine ready");
											break;
										case 1: // SCHEDULED
											MessageBay.displayMessage("Scheduled page load");
											break;
										case 2: // RUNNING
											MessageBay.displayMessage("WebEngine running");

											// Updating the URL bar to display the URL of the page being loaded
											WebBrowserPanel.this.urlField.setText((String) WebEngineGetLocation.invoke(WebViewGetEngine.invoke(WebBrowserPanel.this.webview)));
											break;
										case 3: // SUCCEEDED
											MessageBay.displayMessage("Finished loading page");
											break;
										case 4: // CANCELLED
											MessageBay.displayMessage("Cancelled loading page");
											break;
										case 5: // FAILED
											MessageBay.displayMessage("Failed to load page");
											break;
										}
									}
									System.out.println("\n");
								}
								return null;
							}
				        }));

				// webEngine.getLoadWorker().stateProperty().addListener(
				// new ChangeListener<State>() {
				// public void changed(ObservableValue ov, State oldState, State newState) {
				// if (newState == State.SUCCEEDED) {
				// stage.setTitle(webEngine.getLocation());
				// }
				// }
				// });

				WebEngineLoad.invoke(WebViewGetEngine.invoke(this.webview), url);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		@Override
		public void componentHidden(ComponentEvent e) {
		}

		@Override
		public void componentMoved(ComponentEvent e) {
		}

		@Override
		public void componentResized(ComponentEvent e) {
			// BorderLayout on the JPanel removes the need for explicitly resizing anything here
		}

		@Override
		public void componentShown(ComponentEvent e) {
		}
	}

	public JfxBrowser(Text source, String[] args) {
		// Initial page is either the page stored in the arguments (if there is one stored) or Google
		super(source, new WebBrowserPanel((args != null && args.length > 0) ? args[0] : "http://google.com"), -1, 500, -1, -1, 300, -1);
		_browser = (WebBrowserPanel) _swingComponent;
		_browser.owner = this;
	}

	public void navigate(String url) {
		final String actualURL;
		String urlLower = url.toLowerCase();
		// check if protocol is missing
		if(!(urlLower.startsWith("http://") || url.startsWith("https://") || urlLower.startsWith("ftp://"))) {
			// check if it's a search
			int firstSpace = url.indexOf(" ");
			int firstDot = url.indexOf(".");
			int firstSlash = url.indexOf('/');
			int firstQuestion = url.indexOf('?');
			int firstSQ;
			if(firstSlash == -1) {
				firstSQ = firstQuestion;
			} else if(firstQuestion == -1) {
				firstSQ = firstSlash;
			} else {
				firstSQ = -1;
			}
			if(firstDot <= 0 ||                                        // no '.' or starts with '.' -> search
			        (firstSpace != -1 && firstSpace < firstDot + 1) || // ' ' before '.'            -> search
			        (firstSpace != -1 && firstSpace < firstSQ)) {      // no '/' or '?'             -> search
				// make it a search
				actualURL = UserSettings.searchEngine + url;
			} else {
				// add the missing protocol
				actualURL = "http://" + url;
			}
		} else {
			actualURL = url;
		}
		System.out.println(actualURL);
		try {
			PlatformRunLater.invoke(null, new Runnable() {
				@Override
				public void run() {
					try {
						WebEngineLoad.invoke(WebViewGetEngine.invoke(JfxBrowser.this._browser.webview), actualURL);
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			});
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Navigates JfxBrowser back through history. If end of history reached the user is notified via the MessageBay.
	 * Max size of history is 100 by default.
	 */
	public void navigateBack() {
		try {
			final Object webEngine = WebViewGetEngine.invoke(this._browser.webview);
			try {
				PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						try {
							WebHistoryGo.invoke(WebEngineGetHistory.invoke(webEngine), -1);
						} catch (InvocationTargetException e) {
							MessageBay.displayMessage("Start of History");
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			FreeItems.getInstance().clear();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Navigates JfxBrowser forward through history. If end of history reached the user is notified via the MessageBay.
	 * Max size of history is 100 by default.
	 */
	public void navigateForward() {
		try {
			final Object webEngine = WebViewGetEngine.invoke(this._browser.webview);
			try {
				PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						try {
							WebHistoryGo.invoke(WebEngineGetHistory.invoke(webEngine), 1);
						} catch (InvocationTargetException e) {
							MessageBay.displayMessage("End of History");
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
			FreeItems.getInstance().clear();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Refreshes page in webview.
	 */
	public void refresh() {
		try {
			final Object webEngine = WebViewGetEngine.invoke(this._browser.webview);
			try {
				PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						try {
							WebEngineReload.invoke(webEngine);
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
			MessageBay.displayMessage("Page Refreshed");
			FreeItems.getInstance().clear();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public boolean ItemsLeftClickDropped() {
		Text carried = null;
		if ((carried = FreeItems.getTextAttachedToCursor()) == null) {
			return false;
		}
		
		if(carried.getText().toLowerCase().equals(BACK)) {
			navigateBack();
		} else if(carried.getText().toLowerCase().equals(FORWARD)) {
			navigateForward();
		} else if(carried.getText().toLowerCase().equals(REFRESH)) {
			refresh();
		} else {
			String text = carried.getText().trim();
			this.navigate(text);
			FreeItems.getInstance().clear();
		}
		
		return true;
	}

	@Override
	protected String[] getArgs() {
		String[] r = null;
		if (this._browser.webview != null) {
			try {
				r = new String[] { (String) WebEngineGetLocation.invoke(WebViewGetEngine.invoke(this._browser.webview)) };
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return r;
	}
}
